searchNodes=[{"doc":"Elixir Parser Combinators Author: Matt Mower matt@theartofnavigation.co.uk Version: 0.3.5 Getting Help If you decide to use Ergo and you want help please come find me in the Elixir Discord (where I am sandbags ). Regardless, I would love to hear from you about problems and or suggestions for improvement. What is Ergo? Ergo is an Elixir language parser combinator library. The name 'ergo' means 'therefore' means 'for that reason' which seemed appropriate for a parser. Also note that this is the second attempt I've made at building such a library. The first, Epic, was not completed and, on reflection, not good maintanable code. I did, however, learn a great deal in building it and that learning has put Ergo on a much firmer footing. Using Ergo Installing Add Ergo to your mix.exs file: { :ergo , &quot;~&gt; 0.2&quot; } Then run mix deps . get A Basic Parser What is a parser? As with most parser combinator libraries, Ergo parsers are anonymous functions parameterized by inputs. However Ergo parser functions are wrapped in an Ergo.Parser struct that holds the parser function and can contain additional descriptive metadata (by default a description of the parser behaviour). The Context Ergo parsers operate on a Context struct that gets passed from parser to parser. We can create the context directly: alias Ergo.Context ctx = Context . new ( &quot;string to be parsed&quot; ) Although you will rarely need to do so as there is a helper function Ergo.parse that will do that for you: Ergo . parse ( parser , &quot;string to be parsed&quot; ) Ergo can send debugging information to the Elixir Logger which can be helpful in figuring out why a parser is not working as expected. E.g. Ergo . parse ( parser , &quot;string to be parsed&quot; , debug : true ) The status field of the context will always be either :ok or a tuple of {:error, :error_atom} . The parser can build up a datastructure to return by modifying the ast value of the context. The sequence parser, for example, creates a list of all matching parsers, e.g. Ergo . parse ( sequence ( [ uint ( ) , ignore ( char ( ?, ) ) , uint ( ) ] ) , &quot;1,3&quot; ) =&gt; % Context { status : :ok , ast : [ 1 , 3 ] } The Context in detail At the heart of Ergo is the Context record. All Ergo parser functions take and return a Context . A Context has the following fields: status message input index line col char ast debug status Every parser sets its status to either :ok or a tuple whose first element is the atom :error and whose second element is an atom describing the error, for example: {:error, :unexpected_char} message When a parser is returning the status {:error, _} it can set the message to a user-friendly description of the error. input A binary containing the input remaining to be matched. index The index of the next character to be matched in the input. line The current line of the input the parser has reached. col The current column of the input line that the parser has reached. ast The data structure that the parser constructs from its input. Parsers Ergo comes with a set of basic parsers with which you can assemble your own, more complex, parsers. These are in the form of terminal parsers, combinator parsers, and numeric parsers. Terminal Parsers The terminal parsers are not parameterized by another parser and, in general, are low-level parsers for matching specified characters or literal sequences. eoi() The eoi parser returns :ok if the input is empty otherwise returns {:error, unexpected_eoi} . char(?c) The char parser when given a single character code matches that character. If successful it returns a context with status: :ok with the char and ast parameters set to the character c . Otherwise it returns status: {:error, :unexpected_char} Example: char(?a) matches the character 'a' char(?l..?h) The char parser when given a range of characters. If successful it returns a context with status: :ok with the char and ast parameters set to the matched character. Otherwise it returns status: {:error, :unexpected_char} Example: char(?A..?Z) matches any uppercase letter char(-?c) The char parser when given a negative character matches any character except the specified character and return :ok with the char and ast parameters set to the matched character. Otherwise it returns status: {:error, :unexpected_char} Example: char(-?,) matches any character except a comma char([...]) The char parser when given a list matches a character according to the specifications in the list which can be either ?c, -?c, or [?l..?h]. If successful it returns a context with status: :ok with the char and ast parameters set to the matched character. Otherwise it returns status: {:error, :unexpected_char} Example: char([a..z, 5]) matches any lower case letter or the number 5 digit The digit parser matches a character in the range [0..9] from the input. If successful it returns a context with status: :ok with the char and ast parameters set to the character code of the digit that has been matched. Otherwise it returns status: {:error, :unexpected_char} alpha The alpha parser matches a character in the range [a..z, A..Z] from the input. If successful it returns a context with status: :ok with the char and ast parameters set to the character code of the alpha character that has been matched. Otherwise it returns status: {:error, :unexpected_char} wc The wc parser matches a word character from the input (it is equivalent to the \\w regular expression). If successful it returns a context with status: :ok with the char and ast parameters set to the character code of the alpha character that has been matched. Otherwise it returns status: {:error, :unexpected_char} . ws The ws parser matches a whitespace character from the input (it is equivalent to the \\s regular expression). If successful it returns a context with status: :ok with the char and ast parameters set to the character code of the alpha character that has been matched. Otherwise it returns status: {:error, :unexpected_char} . literal(s) The literal parser is given a binary string and attempts to match it against the input. If successful it returns a context with status: :ok with the char and ast parameters set to the string being matched. Otherwise it returns {:error, :unexpected_char} for the first character that doesn't match the input. Example: literal(&quot;Ergo&quot;) matches the characters 'E', 'r', 'g', and 'o' from the input successively. Numeric parsers These are parsers that build on the terminal parsers to parse diffrent types of numeric values. The utility parsers are technically terminal parsers, since they are not parameterised by a parser, but they are implemented in terms of some of the combinator parsers themselves and hence belong in a separate category. digits The digits parser matches a series of digits from the input. If successful it returns with status: :ok and the ast set to a list of the digit values. uint The uint parser matches a series of digits from the input. If successful it returns with status: :ok and ast parameter set to the integer value of the number matched. decimal The decimal parser matches a series of digits, separated by a single '.' from the input. If successful it returns wtih status: :ok and ast set to the floating point value of the number matched. number The number parser builds upon the uint and decimal parsers to parse any kind of numeric value returning status: :ok and ast set to the integer or floating value of the number parsed. Combinator Parsers The combinator parsers are structural parsers that are parameterised with other parsers to match more complex structures. sequence The sequence/2 parser is used to match a given set of parsers in sequence. Example: p = sequence ( [ literal ( &quot;Hello&quot; ) , char ( ?\\s ) , literal ( &quot;World&quot; ) ] ) p . ( Context . new ( &quot;Hello World&quot; ) ) =&gt; %{ status : :ok , ast : [ &quot;Hello&quot; , &#39; &#39; , &quot;World&quot; ] } The parser p will match first the literal &quot;Hello&quot; then a single space and then the literal &quot;World&quot; . If any of these parsers fail , the sequence parser will fail . This example would more simply be written literal ( &quot;Hello World&quot; ) but more complex examples will use other combinator parsers in the sequence . Optional arguments: ast : fn ast -&gt; ast ctx : fn ctx -&gt; ctx Provide a function that transforms the sequence ast into another form . For example to transform the elements of the list ast into a record type . debug : true | false When the sequence parser runs it logs debugging information . label : &lt;&lt; binary &gt;&gt; Provide a label that can be logged by setting debug : true choice The choice/2 parser is used to match one from a sequence of parsers. It attempts to match the input against each parser in turn. The first parser that matches will cause choice to succeed with the ast set to the ast of the matching parser. If all of the parsers fail to match the input then choice will fail. Signature: choice ( [ parser1 | [ parser2 | ... ] ] , args \\\\ [ ] ) Example: p = choice ( [ literal ( &quot;Hello&quot; ) , literal ( &quot;World&quot; ) ] ) p . ( Context . new ( &quot;Hello World&quot; ) ) -&gt; %{ status : :ok , ast : &quot;Hello&quot; , input : &quot; World&quot; } p . ( Context . new ( &quot;World Hello&quot; ) ) -&gt; %{ status : :ok , ast : &quot;World&quot; , input : &quot; Hello&quot; } boolean = choice ( [ literal ( &quot;true&quot; ) , literal ( &quot;false&quot; ) ] , ast : fn ast -&gt; ast == &quot;true&quot; end ) boolean . ( Context . new ( &quot;true&quot; ) ) -&gt; %{ status : ok , ast : true , input : &quot;&quot; } boolean . ( Context . new ( &quot;false)) -&gt; %{status: ok, ast: false, input: &quot; &quot;} boolean.(Context.new(&quot; Hello ) ) -&gt; %{ status : { :error , ... } , input : &quot;Hello&quot; } many The many/2 parser is used to match another parser repeatedly on the input. In its default form that can be zero times (i.e. no matches) or infinite matches. However using the :min and :max optional arguments allow limits to be specified. Signature: many ( parser , args \\\\ [ ] ) Example: p = many ( char ( ?\\s ) , min : 1 ) p . ( Context . new ( &quot; &quot; ) ) -&gt; %{ status : :ok , input : &quot;&quot; } p = many ( char ( ?\\s ) , min : 2 ) p . ( Context . new ( &quot; Hello&quot; ) ) -&gt; %{ status : { :error , ... } , input : &quot; Hello&quot; } p = many ( char ( ?a .. ?z ) , max : 2 ) p . ( Context . new ( &quot;Hello World)) -&gt; %{status: :ok, ast: [&#39;H&#39;, &#39;e&#39;], input: &quot; llo World &quot;} p = many(char(?a..?z), ast: fn ast -&gt; Enum.count(ast) end) p.(Context.new(&quot; Hello &quot;)) -&gt; %{status: :ok, ast: 5, input: &quot; &quot; } optional The optional/1 parser is used to match another parser on the input and succeeds if it doesn't match or matches once. ignore The ignore/1 parser is designed for use with parsers such as sequence and many . It takes a parser and attempts to match it on the input however, if it succeeds, it returns a nil ast value that parsers can use to ignore it. transform The transform/2 parser is used to change the ast returned by another parser if the other parser is successful. It is useful for modifying the output of parsers that do not directly support a map argument. Signature: squared = transform ( uint ( ) , fn ast -&gt; ast * ast end ) squared . ( Context . new ( &quot;10&quot; ) ) -&gt; %{ status : :ok , ast : 100 } lookahead The lookahead/1 parser attempts to match the parser it is given on the input. If it succeeds it returns with status: :ok but ast: nil and with the input unchanged. Otherwise it returns with status: {:error, :lookahead_fail} . p = lookahead ( literal ( &quot;Hello&quot; ) ) p . ( Context . new ( &quot;Hello World&quot; ) ) -&gt; %{ status : :ok , input : &quot;Hello World&quot; } not_lookahed The not_lookahead/1 parser is the inverse of the looakahead parser in that it attempts to match its parser on the input and where it can do so it returns with status: {:error, :lookahead_fail} . p = not_lookhead ( literal ( &quot;Hello&quot; ) ) p . ( Context . new ( &quot;Hello World&quot; ) ) -&gt; %{ status : { :error , :lookahead_fail } }","ref":"Ergo.html","title":"Ergo","type":"module"},{"doc":"","ref":"Ergo.html#diagnose/3","title":"Ergo.diagnose/3","type":"function"},{"doc":"The parser/2 function is a simple entry point to parsing inputs that constructs the Context record required. Options debug: [true | false] Examples iex&gt; alias Ergo.Terminals iex&gt; parser = Terminals . literal ( &quot;Hello&quot; ) iex&gt; assert % Ergo.Context { status : :ok , ast : &quot;Hello&quot; , input : &quot; World&quot; , index : 5 , line : 1 , col : 6 } = Ergo . parse ( parser , &quot;Hello World&quot; )","ref":"Ergo.html#parse/3","title":"Ergo.parse/3","type":"function"},{"doc":"start/2 should be called before","ref":"Ergo.html#start/2","title":"Ergo.start/2","type":"function"},{"doc":"Ergo.Combinators is the key set of parsers used for combining together other parsers. Parsers choice sequence many optional ignore transform lookeahead not_lookahead","ref":"Ergo.Combinators.html","title":"Ergo.Combinators","type":"module"},{"doc":"The string/1 parser takes a parser that returns an AST which is a string and converts the AST into an atom. Examples iex&gt; alias Ergo iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = many ( wc ( ) ) |&gt; string ( ) |&gt; atom ( ) iex&gt; assert % Context { status : :ok , ast : :fourty_two } = Ergo . parse ( parser , &quot;fourty_two&quot; )","ref":"Ergo.Combinators.html#atom/1","title":"Ergo.Combinators.atom/1","type":"function"},{"doc":"The choice/1 parser takes a list of parsers. It tries each in order attempting to match one. Once a match has been made choice returns the result of the matching parser. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = choice ( [ literal ( &quot;Foo&quot; ) , literal ( &quot;Bar&quot; ) , literal ( &quot;Hello&quot; ) , literal ( &quot;World&quot; ) ] , label : &quot;Foo|Bar|Hello|World&quot; ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; assert % Context { status : :ok , ast : &quot;Hello&quot; , input : &quot; World&quot; , index : 5 , col : 6 } = context iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = choice ( [ literal ( &quot;Foo&quot; ) , literal ( &quot;Bar&quot; ) ] , label : &quot;Foo|Bar&quot; ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; % Context { status : { :error , [ { :no_valid_choice , &quot;Foo|Bar cannot be applied&quot; } ] } , ast : nil , input : &quot;Hello World&quot; } = context","ref":"Ergo.Combinators.html#choice/2","title":"Ergo.Combinators.choice/2","type":"function"},{"doc":"The hoist/1 parser takes a parser expected to return an AST which is a 1-item list. The returned parser extracts the item from the list and returns an AST of just that item. This often comes up with the sequence/2 parser and ignore, where all but one item in a sequence are ignored. Using hoist pulls that item up so that subsequent parsers don't need to deal with the list. Examples iex&gt; alias Ergo iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = sequence ( [ ignore ( many ( char ( ?a ) ) ) , char ( ?b ) ] ) |&gt; hoist ( ) iex&gt; assert % Context { status : :ok , ast : ?b } = Ergo . parse ( parser , &quot;aaaaaaaab&quot; )","ref":"Ergo.Combinators.html#hoist/1","title":"Ergo.Combinators.hoist/1","type":"function"},{"doc":"The ignore/1 parser matches but ignores the AST of its child parser. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = sequence ( [ literal ( &quot;Hello&quot; ) , ignore ( ws ( ) ) , literal ( &quot;World&quot; ) ] ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; assert % Context { status : :ok , ast : [ &quot;Hello&quot; , &quot;World&quot; ] , index : 11 , col : 12 } = context","ref":"Ergo.Combinators.html#ignore/2","title":"Ergo.Combinators.ignore/2","type":"function"},{"doc":"The lazy/1 parser is intended for use in cases where constructing parsers creates a recursive call. By using lazy the original parser call is deferred until later, breaking the infinite recursion.","ref":"Ergo.Combinators.html#lazy/1","title":"Ergo.Combinators.lazy/1","type":"macro"},{"doc":"The lookahead parser accepts a parser and matches it but does not update the context when it succeeds. Example iex&gt; alias Ergo.Context iex&gt; import Ergo . { Combinators , Terminals } iex&gt; parser = lookahead ( literal ( &quot;Hello&quot; ) ) iex&gt; assert % Context { status : :ok , ast : nil , input : &quot;Hello World&quot; , index : 0 } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo . { Combinators , Terminals } iex&gt; parser = lookahead ( literal ( &quot;Helga&quot; ) ) iex&gt; assert % Context { status : { :error , [ { :lookahead_fail , _ } , { :bad_literal , _ } , { :unexpected_char , _ } ] } , index : 3 , col : 4 , input : &quot;lo World&quot; } = Ergo . parse ( parser , &quot;Hello World&quot; )","ref":"Ergo.Combinators.html#lookahead/2","title":"Ergo.Combinators.lookahead/2","type":"function"},{"doc":"Examples This test will need to be rewritten in terms of Ergo . diganose # iex&gt; Logger.disable(self()) # iex&gt; alias Ergo.Context # iex&gt; import Ergo.{Combinators, Terminals} # iex&gt; parser = many(wc(), label: &quot;Chars&quot;) # iex&gt; context = Ergo.parse(parser, &quot;Hello World&quot;) # iex&gt; assert %Context{status: :ok, ast: [?H, ?e, ?l, ?l, ?o], input: &quot; World&quot;, index: 5, col: 6, char: ?o} = context iex&gt; alias Ergo.Context iex&gt; import Ergo . { Combinators , Terminals } iex&gt; parser = many ( wc ( ) , min : 6 ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; assert % Context { status : { :error , [ { :many_less_than_min , &quot;5 &lt; 6&quot; } ] } , ast : nil , input : &quot; World&quot; , index : 5 , col : 6 } = context iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo . { Combinators , Terminals } iex&gt; parser = many ( wc ( ) , max : 3 ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; assert % Context { status : :ok , ast : [ ?H , ?e , ?l ] , input : &quot;lo World&quot; , index : 3 , col : 4 } = context iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo . { Combinators , Terminals } iex&gt; parser = many ( wc ( ) , ast : &amp; Enum . count / 1 ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; assert % Context { status : :ok , ast : 5 , input : &quot; World&quot; , index : 5 , col : 6 } = context","ref":"Ergo.Combinators.html#many/2","title":"Ergo.Combinators.many/2","type":"function"},{"doc":"A ctx: function should be passed &amp; return the whole context. It takes precendence over an ast: function that receives and returns a modified AST. Otherwise the identity function is returned. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Combinators iex&gt; f = mapping_fn ( ctx : fn _ -&gt; :kazam end ) iex&gt; assert :kazam = f . ( % Context { } )","ref":"Ergo.Combinators.html#mapping_fn/1","title":"Ergo.Combinators.mapping_fn/1","type":"function"},{"doc":"The not_lookahead parser accepts a parser and attempts to match it. If the match fails the not_lookahead parser returns status: :ok but does not affect the context otherwise. If the match succeeds the not_lookahead parser fails with {:error, :lookahead_fail} Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.{Combinators, Terminals} iex&gt; parser = not_lookahead(literal(&quot;Foo&quot;)) iex&gt; assert %Context{status: :ok, input: &quot;Hello World&quot;} = Ergo.parse(parser, &quot;Hello World&quot;) iex&gt; alias Ergo.{Context, Parser} iex&gt; import Ergo.{Combinators, Terminals} iex&gt; parser = not_lookahead(literal(&quot;Hello&quot;)) iex&gt; assert %Context{status: {:error, [{:lookahead_fail, &quot;Satisfied: literal&lt;Hello&gt;&quot;}]}, input: &quot;Hello World&quot;} = Ergo.parse(parser, &quot;Hello World&quot;)","ref":"Ergo.Combinators.html#not_lookahead/2","title":"Ergo.Combinators.not_lookahead/2","type":"function"},{"doc":"Examples iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; context = Ergo . parse ( optional ( literal ( &quot;Hello&quot; ) ) , &quot;Hello World&quot; ) iex&gt; assert % Context { status : :ok , ast : &quot;Hello&quot; , input : &quot; World&quot; , index : 5 , col : 6 } = context In this example we deliberately ensure that the Context ast is not nil iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo . { Terminals , Combinators } iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot; World&quot; , ast : [ ] ) iex&gt; parser = optional ( literal ( &quot;Hello&quot; ) ) iex&gt; new_context = Parser . invoke ( parser , context ) iex&gt; assert % Context { status : :ok , ast : nil , input : &quot; World&quot; , index : 0 , col : 1 } = new_context","ref":"Ergo.Combinators.html#optional/2","title":"Ergo.Combinators.optional/2","type":"function"},{"doc":"","ref":"Ergo.Combinators.html#parse_many/5","title":"Ergo.Combinators.parse_many/5","type":"function"},{"doc":"","ref":"Ergo.Combinators.html#parser_labels/1","title":"Ergo.Combinators.parser_labels/1","type":"function"},{"doc":"The replace/3 combinator replaces the AST value of it's child with a constant. Examples iex&gt; alias Ergo.Context iex&gt; alias Ergo iex&gt; import Ergo . { Combinators , Terminals } iex&gt; parser = ignore ( literal ( &quot;foo&quot; ) ) |&gt; replace ( :foo ) iex&gt; assert % Context { status : { :error , _ } } = Ergo . parse ( parser , &quot;flush&quot; )","ref":"Ergo.Combinators.html#replace/3","title":"Ergo.Combinators.replace/3","type":"function"},{"doc":"The satisfy/3 parser takes a parser and a predicate function. If the parser is successful the AST is passed to the predicate function. If the predicate function returns true the parser returns the successful context, otherwise an error context is returned. Example iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators , Numeric } iex&gt; parser = satisfy ( any ( ) , fn char -&gt; char in ( ?0 .. ?9 ) end , label : &quot;digit char&quot; ) iex&gt; assert % Context { status : :ok , ast : ?4 } = Ergo . parse ( parser , &quot;4&quot; ) iex&gt; assert % Context { status : { :error , [ { :unsatisfied , &quot;Failed to satisfy: digit char&quot; } ] } } = Ergo . parse ( parser , &quot;!&quot; ) iex&gt; parser = satisfy ( number ( ) , fn n -&gt; Integer . mod ( n , 2 ) == 0 end , label : &quot;even number&quot; ) iex&gt; assert % Context { status : :ok , ast : 42 } = Ergo . diagnose ( parser , &quot;42&quot; ) iex&gt; assert % Context { status : { :error , [ { :unsatisfied , &quot;Failed to satisfy: even number&quot; } ] } } = Ergo . parse ( parser , &quot;27&quot; )","ref":"Ergo.Combinators.html#satisfy/3","title":"Ergo.Combinators.satisfy/3","type":"function"},{"doc":"Examples iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = sequence ( [ literal ( &quot;Hello&quot; ) , ws ( ) , literal ( &quot;World&quot; ) ] ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) % Context { status : :ok , ast : [ &quot;Hello&quot; , ?\\s , &quot;World&quot; ] , index : 11 , line : 1 , col : 12 } = context This test will need to be rewritten in terms of Ergo . diagnose # iex&gt; Logger.disable(self()) # iex&gt; alias Ergo.Context # iex&gt; import Ergo.{Terminals, Combinators} # iex&gt; parser = sequence([literal(&quot;Hello&quot;), ws(), literal(&quot;World&quot;)], label: &quot;HelloWorld&quot;) # iex&gt; context = Ergo.parse(parser, &quot;Hello World&quot;) # iex&gt; assert %Context{status: :ok, ast: [&quot;Hello&quot;, ?\\s, &quot;World&quot;], index: 11, line: 1, col: 12} = context iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = sequence ( [ literal ( &quot;Hello&quot; ) , ws ( ) , literal ( &quot;World&quot; ) ] , ast : fn ast -&gt; Enum . join ( ast , &quot; &quot; ) end ) iex&gt; context = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; assert % Context { status : :ok , ast : &quot;Hello 32 World&quot; , index : 11 , line : 1 , col : 12 } = context This test will need to be rewritten in terms of Ergo . diagnose # iex&gt; Logger.disable(self()) # iex&gt; alias Ergo.Context # iex&gt; import Ergo.{Terminals, Combinators} # iex&gt; parser = sequence([literal(&quot;Hello&quot;), ws(), literal(&quot;World&quot;)], label: &quot;HelloWorld&quot;, ast: fn ast -&gt; Enum.join(ast, &quot; &quot;) end) # iex&gt; context = Ergo.parse(parser, &quot;Hello World&quot;) # iex&gt; assert %Context{status: :ok, ast: &quot;Hello 32 World&quot;, index: 11, line: 1, col: 12} = context iex&gt; alias Ergo.Context iex&gt; import Ergo . { Combinators , Terminals } iex&gt; parser = sequence ( [ literal ( &quot;foo&quot; ) , ws ( ) , literal ( &quot;bar&quot; ) ] ) iex&gt; assert % Context { status : { :error , [ { :bad_literal , _ } , { :unexpected_char , _ } ] } } = Ergo . parse ( parser , &quot;Hello World&quot; )","ref":"Ergo.Combinators.html#sequence/2","title":"Ergo.Combinators.sequence/2","type":"function"},{"doc":"The string/1 parser takes a parser that returns an AST which is a list of characters and converts the AST into a string. Examples iex&gt; alias Ergo iex&gt; alias Ergo.Context iex&gt; import Ergo . { Terminals , Combinators } iex&gt; parser = many ( alpha ( ) ) |&gt; string ( ) iex&gt; assert % Context { status : :ok , ast : &quot;FourtyTwo&quot; } = Ergo . parse ( parser , &quot;FourtyTwo&quot; )","ref":"Ergo.Combinators.html#string/1","title":"Ergo.Combinators.string/1","type":"function"},{"doc":"The transform/2 parser runs a transforming function on the AST of its child parser. Examples # Sum the digits iex&gt; alias Ergo.Context iex&gt; import Ergo . { Combinators , Terminals } iex&gt; digit_to_int = fn d -&gt; List . to_string ( [ d ] ) |&gt; String . to_integer ( ) end iex&gt; t_fn = fn ast -&gt; ast |&gt; Enum . map ( digit_to_int ) |&gt; Enum . sum ( ) end iex&gt; parser = sequence ( [ digit ( ) , digit ( ) , digit ( ) , digit ( ) ] ) |&gt; transform ( t_fn ) iex&gt; context = Ergo . parse ( parser , &quot;1234&quot; ) iex&gt; % Context { status : :ok , ast : 10 , index : 4 , line : 1 , col : 5 } = context","ref":"Ergo.Combinators.html#transform/3","title":"Ergo.Combinators.transform/3","type":"function"},{"doc":"Ergo.Context defines the Context struct that is used to maintain parser state as the various parsers work, and various functions for creating &amp; manipulating contexts. Fields invoke_fn The invoke_fn defines the parsing entry point used for calling parsers. status When a parser returns it either sets status to :ok to indicate that it was successful or to a tuple {:error, :error_atom} where error_atom is an atom indiciting the specific type of error. It may optionally set the message field to a human readable message. input The binary input being parsed. index Represents the position in the input which has been read so far. Initially 0 it increments for each character processed. line Represents the current line of the input. Initially 1 it increments whenever a is read from the input. col Represents the current column of the input. Initially 1 it is incremented every time a character is read from the input and automatically resets whenever a is read. ast Represents the current data structure being built from the input. tracks Parsers for which track: true is specified (by default this is most of the combinator parsers but not the terminal parsers) will add themselves to the Context tracks in the form of {ref, index} . If the same parser attempts to add itself a second time at the same index an error is thrown because a cycle has been detected.","ref":"Ergo.Context.html","title":"Ergo.Context","type":"module"},{"doc":"Examples iex&gt; alias Ergo.Context iex&gt; context = ...&gt; Context . new ( &amp; Ergo.Parser . call / 2 , &quot;Hello World&quot; ) ...&gt; |&gt; Context . add_error ( :unexpected_char , &quot;Expected &#39;e&#39; got &#39;.&#39;&quot; ) iex&gt; assert is_nil ( context . ast ) iex&gt; assert { :error , [ { :unexpected_char , &quot;Expected &#39;e&#39; got &#39;.&#39;&quot; } ] } = context . status iex&gt; alias Ergo.Context iex&gt; context = ...&gt; Context . new ( &amp; Ergo.Parser . call / 2 , &quot;Hello World&quot; ) ...&gt; |&gt; Context . add_error ( :unexpected_char , &quot;Expected &#39;e&#39; got &#39;.&#39;&quot; ) ...&gt; |&gt; Context . add_error ( :literal_failed , &quot;Expected &#39;end&#39;&quot; ) iex&gt; assert is_nil ( context . ast ) iex&gt; assert { :error , [ { :literal_failed , &quot;Expected &#39;end&#39;&quot; } , { :unexpected_char , &quot;Expected &#39;e&#39; got &#39;.&#39;&quot; } ] } = context . status","ref":"Ergo.Context.html#add_error/3","title":"Ergo.Context.add_error/3","type":"function"},{"doc":"","ref":"Ergo.Context.html#add_errors/2","title":"Ergo.Context.add_errors/2","type":"function"},{"doc":"Because we build ASTs using lists they end up in reverse order. This method reverses the AST back to in-parse-order Examples iex&gt; context = Ergo.Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; , ast : [ 4 , 3 , 2 , 1 ] ) iex&gt; assert % Context { ast : [ 1 , 2 , 3 , 4 ] } = Context . ast_in_parsed_order ( context )","ref":"Ergo.Context.html#ast_in_parsed_order/1","title":"Ergo.Context.ast_in_parsed_order/1","type":"function"},{"doc":"Where an AST has been built from individual characters and needs to be converted to a string Examples iex&gt; context = Ergo.Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; , ast : [ ?H , ?e , ?l , ?l , ?o ] ) iex&gt; assert % Context { ast : &quot;Hello&quot; } = Context . ast_to_string ( context )","ref":"Ergo.Context.html#ast_to_string/1","title":"Ergo.Context.ast_to_string/1","type":"function"},{"doc":"Called to perform an arbitrary transformation on the AST value of a Context. Examples iex&gt; alias Ergo.Context iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; , ast : &quot;Hello World&quot; ) iex&gt; assert % Context { ast : &quot;Hello World&quot; } = Context . ast_transform ( context , &amp; Function . identity / 1 ) iex&gt; alias Ergo.Context iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; , ast : &quot;Hello World&quot; ) iex&gt; assert % Context { ast : 11 } = Context . ast_transform ( context , &amp; String . length / 1 ) iex&gt; alias Ergo.Context iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; , ast : &quot;Hello World&quot; ) iex&gt; assert % Context { ast : &quot;Hello World&quot; } = Context . ast_transform ( context , nil )","ref":"Ergo.Context.html#ast_transform/2","title":"Ergo.Context.ast_transform/2","type":"function"},{"doc":"The ignore parser matches but returns a nil for the AST. Parsers like sequence accumulate these nil values. Call this function to remove them Examples iex&gt; context = Ergo.Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; , ast : [ &quot;Hello&quot; , nil , &quot;World&quot; , nil ] ) iex&gt; assert % Context { ast : [ &quot;Hello&quot; , &quot;World&quot; ] } = Context . ast_without_ignored ( context )","ref":"Ergo.Context.html#ast_without_ignored/1","title":"Ergo.Context.ast_without_ignored/1","type":"function"},{"doc":"","ref":"Ergo.Context.html#clip/2","title":"Ergo.Context.clip/2","type":"function"},{"doc":"","ref":"Ergo.Context.html#dec_depth/1","title":"Ergo.Context.dec_depth/1","type":"function"},{"doc":"","ref":"Ergo.Context.html#inc_depth/1","title":"Ergo.Context.inc_depth/1","type":"function"},{"doc":"new returns a newly initialised Context with input set to the string passed in. Examples: iex&gt; Context.new(&amp;Ergo.Parser.call/2, &quot;Hello World&quot;) %Context{status: :ok, input: &quot;Hello World&quot;, line: 1, col: 1, index: 0, tracks: %MapSet{}, invoke_fn: &amp;Ergo.Parser.call/2}","ref":"Ergo.Context.html#new/3","title":"Ergo.Context.new/3","type":"function"},{"doc":"Reads the next character from the input of the passed in Context . If the input is empty returns status: {:error, :unexpected_eoi} . Otherwise returns a new Context setting ast to the character read and incrementing positional variables such as index , line , and column appropriately. Examples iex&gt; context = Context.next_char(Context.new(&amp;Ergo.Parser.call/2, &quot;&quot;)) iex&gt; assert %Context{status: {:error, [{:unexpected_eoi, &quot;Unexpected end of input&quot;}] }} = context iex&gt; context = Context.next_char(Context.new(&amp;Ergo.Parser.call/2, &quot;Hello World&quot;)) iex&gt; assert %Context{status: :ok, input: &quot;ello World&quot;, ast: ?H, index: 1, line: 1, col: 2} = context","ref":"Ergo.Context.html#next_char/1","title":"Ergo.Context.next_char/1","type":"function"},{"doc":"Returns truthy value if the parser referred to by ref has already been called for the index index . Examples iex&gt; alias Ergo.Context iex&gt; parser_ref = 123 iex&gt; context = Context.new(&amp;Ergo.Parser.call/2, &quot;Hello World&quot;) |&gt; Context.track_parser(parser_ref) iex&gt; assert Context.parser_tracked?(context, parser_ref)","ref":"Ergo.Context.html#parser_tracked?/2","title":"Ergo.Context.parser_tracked?/2","type":"function"},{"doc":"Examples iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;Hello&quot; ) iex&gt; assert % Context { status : :ok , ast : ?H , input : &quot;ello&quot; , index : 1 , line : 1 , col : 2 } = Context . peek ( context ) iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; ) iex&gt; assert % Context { status : { :error , [ { :unexpected_eoi , &quot;Unexpected end of input&quot; } ] } , index : 0 , line : 1 , col : 1 } = Context . peek ( context )","ref":"Ergo.Context.html#peek/1","title":"Ergo.Context.peek/1","type":"function"},{"doc":"Clears the value of the status and ast fields to ensure that the wrong status information cannot be returned from a child parser. Examples iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;Hello World&quot; ) iex&gt; context = %{ context | status : { :error , [ { :inexplicable_error , &quot;What the…&quot; } ] } , ast : true } iex&gt; context = Context . reset_status ( context ) iex&gt; assert % Context { status : :ok , ast : nil } = context","ref":"Ergo.Context.html#reset_status/1","title":"Ergo.Context.reset_status/1","type":"function"},{"doc":"","ref":"Ergo.Context.html#trace/3","title":"Ergo.Context.trace/3","type":"function"},{"doc":"","ref":"Ergo.Context.html#trace_match/4","title":"Ergo.Context.trace_match/4","type":"function"},{"doc":"Updates the Context to track that the parser referred to by ref has been called for the index index . Examples: iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;Hello World&quot; ) iex&gt; parser = literal ( &quot;Hello&quot; ) iex&gt; context2 = Context . track_parser ( context , parser . ref ) iex&gt; assert MapSet . member? ( context2 . tracks , { parser . ref , 0 } )","ref":"Ergo.Context.html#track_parser/2","title":"Ergo.Context.track_parser/2","type":"function"},{"doc":"","ref":"Ergo.Context.html#transform/2","title":"Ergo.Context.transform/2","type":"function"},{"doc":"The Meta parsers are not really parsers at all but operate within the parsing framework. All Meta parsers are combinators that accept a parser and either report on or modify its operation.","ref":"Ergo.Meta.html","title":"Ergo.Meta","type":"module"},{"doc":"The around parser is a pass-thru parser that can be used to run code before or after another parser. For example to inspect the context or output additional debugging information. Specify on of around(parser, before: fn ctx -&gt; … end) around(parser, after: fn ctx, parsed_ctx -&gt; … end) around(parser, before: fn ctx -&gt; … end, after: fn ctx, parsed_ctx -&gt; … end) Note that where both before: and after: are specified, the after function will receive both the context before and after parsing as separate parameters. If neither of before: nor around: are specified an error will be raised. Examples iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo.Meta iex&gt; null_parser = Parser . combinator ( &quot;null_parser&quot; , fn % Context { } = ctx -&gt; ctx end ) iex&gt; assert_raise ( RuntimeError , fn -&gt; Ergo . parse ( around ( null_parser ) , &quot;&quot; ) end ) iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo.Meta iex&gt; null_parser = Parser . combinator ( &quot;null_parser&quot; , fn % Context { } = ctx -&gt; ctx end ) iex&gt; parser = around ( null_parser , before : fn _ctx -&gt; send ( self ( ) , :before ) end ) iex&gt; Ergo . parse ( parser , &quot;&quot; ) iex&gt; assert_receive :before iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo.Meta iex&gt; null_parser = Parser . combinator ( &quot;null_parser&quot; , fn % Context { } = ctx -&gt; ctx end ) iex&gt; parser = around ( null_parser , after : fn _ctx , _new_ctx -&gt; send ( self ( ) , :after ) end ) iex&gt; Ergo . parse ( parser , &quot;&quot; ) iex&gt; assert_receive :after iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo.Meta iex&gt; null_parser = Parser . combinator ( &quot;null_parser&quot; , fn % Context { } = ctx -&gt; ctx end ) iex&gt; parser = around ( null_parser , before : fn _ctx -&gt; send ( self ( ) , :before ) end , after : fn _ctx , _new_ctx -&gt; send ( self ( ) , :after ) end ) iex&gt; Ergo . parse ( parser , &quot;&quot; ) iex&gt; assert_receive :before iex&gt; assert_receive :after","ref":"Ergo.Meta.html#around/2","title":"Ergo.Meta.around/2","type":"function"},{"doc":"The failed parser is a combinator that invokes a parser and if the parser fails runs the given function on the resulting context. This can be used to output additional debugging information where failure was unexpected. Examples iex&gt; alias Ergo.Parser iex&gt; import Ergo.Meta iex&gt; failing_parser = Parser . combinator ( &quot;err_parser&quot; , fn ctx -&gt; %{ ctx | status : { :error , :unfathomable_error } } end ) iex&gt; parser = failed ( failing_parser , fn _ctx -&gt; send ( self ( ) , :failed ) end ) iex&gt; Ergo . parse ( parser , &quot;&quot; ) iex&gt; assert_received :failed","ref":"Ergo.Meta.html#failed/3","title":"Ergo.Meta.failed/3","type":"function"},{"doc":"","ref":"Ergo.Meta.html#suppress_caller_logging/1","title":"Ergo.Meta.suppress_caller_logging/1","type":"function"},{"doc":"The Parsers module exists to house utility parsers that while they are terminals in the sense that they are not parameterised, they internally make use of parsers from the Combinators module. Parsers uint decimal digits","ref":"Ergo.Numeric.html","title":"Ergo.Numeric","type":"module"},{"doc":"Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Numeric iex&gt; context = Ergo . parse ( decimal ( ) , &quot;234.56&quot; ) iex&gt; assert % Context { status : :ok , ast : 234.56 } = context","ref":"Ergo.Numeric.html#decimal/1","title":"Ergo.Numeric.decimal/1","type":"function"},{"doc":"The digits parser matches a series of at least one digit and returns an enumeration of the digits. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Numeric iex&gt; context = Ergo . parse ( digits ( ) , &quot;2345&quot; ) iex&gt; assert % Context { status : :ok , ast : [ 2 , 3 , 4 , 5 ] } = context","ref":"Ergo.Numeric.html#digits/1","title":"Ergo.Numeric.digits/1","type":"function"},{"doc":"The number parser matches both integer and decimal string and converts them into their appropriate Elixir integer or float values. Examples iex&gt; import Ergo.Numeric iex&gt; assert %{ status : :ok , ast : 42 } = Ergo . parse ( number ( ) , &quot;42&quot; ) iex&gt; import Ergo.Numeric iex&gt; assert %{ status : :ok , ast : - 42 } = Ergo . parse ( number ( ) , &quot;-42&quot; ) iex&gt; import Ergo.Numeric iex&gt; assert %{ status : :ok , ast : 42.0 } = Ergo . parse ( number ( ) , &quot;42.0&quot; ) iex&gt; import Ergo.Numeric iex&gt; assert %{ status : :ok , ast : - 42.0 } = Ergo . parse ( number ( ) , &quot;-42.0&quot; ) iex&gt; import Ergo.Numeric iex&gt; assert %{ status : :ok , ast : 0 } = Ergo . parse ( number ( ) , &quot;0&quot; ) iex&gt; import Ergo.Numeric iex&gt; assert %{ status : :ok , ast : 0 } = Ergo . parse ( number ( ) , &quot;0000&quot; ) iex&gt; import Ergo.Numeric iex&gt; assert %{ status : { :error , _ } } = Ergo . parse ( number ( ) , &quot;Fourty Two&quot; ) iex&gt; import Ergo.Numeric iex&gt; assert %{ status : :ok , ast : 42 } = Ergo . parse ( number ( ) , &quot;42Fourty Two&quot; )","ref":"Ergo.Numeric.html#number/1","title":"Ergo.Numeric.number/1","type":"function"},{"doc":"The unit parser matches a series of at least one digit and returns the integer value of the digits. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Numeric iex&gt; context = Ergo . parse ( uint ( ) , &quot;2345&quot; ) iex&gt; assert % Context { status : :ok , ast : 2345 , index : 4 , col : 5 } = context","ref":"Ergo.Numeric.html#uint/1","title":"Ergo.Numeric.uint/1","type":"function"},{"doc":"Ergo.Parser contains the Parser record type. Ergo parsers are anonymous functions but we embed them in a Parser record that can hold arbitrary metadata. The primary use for the metadata is the storage of debugging information.","ref":"Ergo.Parser.html","title":"Ergo.Parser","type":"module"},{"doc":"call/2 invokes the specified parser by calling its parsing function with the specified context having first reset the context status.","ref":"Ergo.Parser.html#call/2","title":"Ergo.Parser.call/2","type":"function"},{"doc":"Create a new combinator parser with the given label, parsing function, and optional metadata.","ref":"Ergo.Parser.html#combinator/3","title":"Ergo.Parser.combinator/3","type":"function"},{"doc":"diagnose/2 invokes the specified parser by calling its parsing function with the specific context while tracking the progress of the parser. The progress can be retrieved from the progress key of the returned context. Examples iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo . { Combinators , Terminals } iex&gt; context = Context . new ( &amp; Ergo.Parser . diagnose / 2 , &quot;Hello World&quot; ) iex&gt; parser = many ( wc ( ) ) iex&gt; assert %{ status : :ok } = Parser . invoke ( parser , context )","ref":"Ergo.Parser.html#diagnose/2","title":"Ergo.Parser.diagnose/2","type":"function"},{"doc":"","ref":"Ergo.Parser.html#enter/1","title":"Ergo.Parser.enter/1","type":"function"},{"doc":"invoke/2 is the main entry point for the parsing process. It looks up the parser control function within the Context and uses it to run the given parser . This indirection allows a different control function to be specified, e.g. by the diagnose entry point which can wrap the parser call, while still calling the same parsing function (i.e. we are not introducing debugging variants of the parsers that could be subject to different behaviours)","ref":"Ergo.Parser.html#invoke/2","title":"Ergo.Parser.invoke/2","type":"function"},{"doc":"","ref":"Ergo.Parser.html#leave/1","title":"Ergo.Parser.leave/1","type":"function"},{"doc":"","ref":"Ergo.Parser.html#process/2","title":"Ergo.Parser.process/2","type":"function"},{"doc":"The rewrite_error/2 call allows a higher-level parser to rewrite the error returned by a subordinate parser, translating it into something a user is more likely to be able to understand.","ref":"Ergo.Parser.html#rewrite_error/2","title":"Ergo.Parser.rewrite_error/2","type":"function"},{"doc":"","ref":"Ergo.Parser.html#should_debug?/2","title":"Ergo.Parser.should_debug?/2","type":"function"},{"doc":"Create a new terminal parser with the given label, parsing function, and optional metadata.","ref":"Ergo.Parser.html#terminal/3","title":"Ergo.Parser.terminal/3","type":"function"},{"doc":"","ref":"Ergo.Parser.html#trace_in/3","title":"Ergo.Parser.trace_in/3","type":"function"},{"doc":"","ref":"Ergo.Parser.html#trace_out/3","title":"Ergo.Parser.trace_out/3","type":"function"},{"doc":"track_parser first checks if the parser has already been tracked for the current input index and, if it has, raises a CycleError to indicate the parser is in a loop. Otherwise it adds the parser at the current index. Examples iex&gt; alias Ergo.{Context, Parser} iex&gt; import Ergo.{Terminals, Combinators} iex&gt; context = Context.new(&amp;Ergo.Parser.call/2, &quot;Hello World&quot;) iex&gt; parser = many(char(?H)) iex&gt; context2 = Parser.track_parser(context, parser) iex&gt; assert Context.parser_tracked?(context2, parser.ref)","ref":"Ergo.Parser.html#track_parser/2","title":"Ergo.Parser.track_parser/2","type":"function"},{"doc":"","ref":"Ergo.ParserRefs.html","title":"Ergo.ParserRefs","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Ergo.ParserRefs.html#child_spec/1","title":"Ergo.ParserRefs.child_spec/1","type":"function"},{"doc":"Returns the next parser ref id, a monotonically increasing value starting at 0 Examples iex&gt; assert is_integer ( Ergo.ParserRefs . next_ref ( ) )","ref":"Ergo.ParserRefs.html#next_ref/0","title":"Ergo.ParserRefs.next_ref/0","type":"function"},{"doc":"Starts the refs agent. There is no need to call this directly as it will be started as part of the calling applications supervision tree via mix.exs","ref":"Ergo.ParserRefs.html#start_link/1","title":"Ergo.ParserRefs.start_link/1","type":"function"},{"doc":"Ergo.Terminals contains the terminal parsers, which are those parsers not parameterized with other parsers and therefore work more at the level of text than structure.","ref":"Ergo.Terminals.html","title":"Ergo.Terminals","type":"module"},{"doc":"The alpha/0 parser accepts a single character in the range a..z or A..Z. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = alpha ( ) iex&gt; assert % Context { status : :ok , input : &quot;ello World&quot; , ast : ?H , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = alpha ( ) iex&gt; assert % Context { status : :ok , input : &quot;llo World&quot; , ast : ?e , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot;ello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = alpha ( ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Expected: [a..z, A..Z] Actual: &quot; } ] } , input : &quot; World&quot; } = Ergo . parse ( parser , &quot; World&quot; )","ref":"Ergo.Terminals.html#alpha/0","title":"Ergo.Terminals.alpha/0","type":"function"},{"doc":"The any/0 parser matches any character. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = any ( ) iex&gt; assert % Context { status : :ok , ast : ?H } = Ergo . parse ( parser , &quot;H&quot; ) iex&gt; assert % Context { status : :ok , ast : ?e } = Ergo . parse ( parser , &quot;e&quot; ) iex&gt; assert % Context { status : :ok , ast : ?! } = Ergo . parse ( parser , &quot;!&quot; ) iex&gt; assert % Context { status : :ok , ast : ?0 } = Ergo . parse ( parser , &quot;0&quot; )","ref":"Ergo.Terminals.html#any/0","title":"Ergo.Terminals.any/0","type":"function"},{"doc":"The char/1 parser is a terminal parser that matches a specific character. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( ?H ) iex&gt; assert % Context { status : :ok , ast : ?H , input : &quot;ello World&quot; , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( ?h ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Expected: h Actual: H&quot; } ] } , input : &quot;Hello World&quot; } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( ?H ) iex&gt; assert % Context { status : { :error , [ { :unexpected_eoi , &quot;Unexpected end of input&quot; } ] } } = Ergo . parse ( parser , &quot;&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( ?A .. ?Z ) iex&gt; assert % Context { status : :ok , ast : ?H , input : &quot;ello World&quot; , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( ?a .. ?z ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Expected: a..z Actual: H&quot; } ] } , input : &quot;Hello World&quot; } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( ?A .. ?Z ) iex&gt; assert % Context { status : { :error , [ { :unexpected_eoi , &quot;Unexpected end of input&quot; } ] } } = Ergo . parse ( parser , &quot;&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( [ ?a .. ?z , ?A .. ?Z ] ) iex&gt; assert % Context { status : :ok , ast : ?H , input : &quot;ello World&quot; , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( [ ?a .. ?z , ?A .. ?Z ] ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Expected: [a..z, A..Z] Actual: 0&quot; } ] } , input : &quot;0000&quot; } = Ergo . parse ( parser , &quot;0000&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( - ?0 ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Should not have matched 0&quot; } ] } , input : &quot;0000&quot; } = Ergo . parse ( parser , &quot;0000&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = char ( - ?a ) iex&gt; assert % Context { status : :ok , input : &quot;000&quot; , ast : ?0 , index : 1 , col : 2 } = Ergo . parse ( parser , &quot;0000&quot; )","ref":"Ergo.Terminals.html#char/2","title":"Ergo.Terminals.char/2","type":"function"},{"doc":"The delimited_text/2 parser matches a sequence of text delimited open_char and close_char . Because it is expected that open_char may appear multiple times within the sequence it balances the tokens to ensure the right number of closing tokens is matched. Examples iex&gt; alias Ergo iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = delimited_text ( ?{ , ?} ) iex&gt; assert % Context { status : :ok , ast : &quot;{return {foo: \\&quot; bar \\&quot; , bar: {baz: \\&quot; quux \\&quot; }};}&quot; , input : &quot;&quot; } = Ergo . parse ( parser , &quot;{return {foo: \\&quot; bar \\&quot; , bar: {baz: \\&quot; quux \\&quot; }};}&quot; ) iex&gt; assert % Context { status : :ok , ast : &quot;{function b(y) {return x + y;}; return b;}&quot; , input : &quot;foo&quot; } = Ergo . parse ( parser , &quot;{function b(y) {return x + y;}; return b;}foo&quot; )","ref":"Ergo.Terminals.html#delimited_text/3","title":"Ergo.Terminals.delimited_text/3","type":"function"},{"doc":"The digit/0 parser accepts a character in the range of 0..9 Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = digit ( ) iex&gt; assert % Context { status : :ok , ast : ?0 , input : &quot;000&quot; , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot;0000&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; import Ergo.Terminals iex&gt; parser = digit ( ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Expected: 0..9 Actual: A&quot; } ] } , input : &quot;AAAA&quot; , index : 0 , line : 1 , col : 1 } = Ergo . parse ( parser , &quot;AAAA&quot; ) iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo.Terminals iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; ) iex&gt; parser = digit ( ) iex&gt; assert % Context { status : { :error , [ { :unexpected_eoi , &quot;Unexpected end of input&quot; } ] } , input : &quot;&quot; , index : 0 , line : 1 , col : 1 } = Parser . invoke ( parser , context )","ref":"Ergo.Terminals.html#digit/0","title":"Ergo.Terminals.digit/0","type":"function"},{"doc":"The eoi parser is a terminal parser that checks whether the input has been fully consumed. If there is input remaining to be parsed the return context status is set to :error. Examples iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo.Terminals iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;&quot; ) iex&gt; assert % Context { status : :ok , ast : nil } = Parser . invoke ( eoi ( ) , context ) iex&gt; alias Ergo . { Context , Parser } iex&gt; import Ergo.Terminals iex&gt; context = Context . new ( &amp; Ergo.Parser . call / 2 , &quot;Hello World&quot; ) iex&gt; assert % Context { status : { :error , [ { :not_eoi , &quot;Input not empty: Hello World&quot; } ] } , input : &quot;Hello World&quot; } = Parser . invoke ( eoi ( ) , context )","ref":"Ergo.Terminals.html#eoi/0","title":"Ergo.Terminals.eoi/0","type":"function"},{"doc":"The literal/1 parser matches the specified string character by character. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = literal ( &quot;Hello&quot; ) iex&gt; assert % Context { status : :ok , input : &quot; World&quot; , ast : &quot;Hello&quot; , index : 5 , line : 1 , col : 6 } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = literal ( &quot;Hellx&quot; ) iex&gt; assert % Context { status : { :error , [ { :bad_literal , &quot;literal&lt;Hellx&gt;&quot; } , { :unexpected_char , &quot;Expected: x Actual: o&quot; } ] } , input : &quot;o World&quot; , index : 4 , line : 1 , col : 5 } = Ergo . parse ( parser , &quot;Hello World&quot; )","ref":"Ergo.Terminals.html#literal/2","title":"Ergo.Terminals.literal/2","type":"function"},{"doc":"The not_char matcher accepts a char or a list of chars and will match any char that is not in the list. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = not_char ( ?0 ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Should not have matched 0&quot; } ] } , input : &quot;0000&quot; } = Ergo . parse ( parser , &quot;0000&quot; ) iex&gt; assert % Context { status : :ok , ast : ?1 } = Ergo . parse ( parser , &quot;1111&quot; ) iex&gt; parser = not_char ( [ ?{ , ?} ] ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Should not have matched {&quot; } ] } , input : &quot;{}&quot; } = Ergo . parse ( parser , &quot;{}&quot; ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Should not have matched }&quot; } ] } , input : &quot;}&quot; } = Ergo . parse ( parser , &quot;}&quot; )","ref":"Ergo.Terminals.html#not_char/2","title":"Ergo.Terminals.not_char/2","type":"function"},{"doc":"The wc/0 parser parses a word character and is analagous to the \\w regular expression. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = wc ( ) iex&gt; assert % Context { status : :ok , ast : ?H , input : &quot;ello World&quot; , index : 1 , col : 2 } = Ergo . parse ( parser , &quot;Hello World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = wc ( ) iex&gt; assert % Context { status : :ok , ast : ?0 , input : &quot; World&quot; , index : 1 , col : 2 } = Ergo . parse ( parser , &quot;0 World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = wc ( ) iex&gt; assert % Context { status : :ok , ast : ?_ , input : &quot;Hello&quot; , index : 1 , col : 2 } = Ergo . parse ( parser , &quot;_Hello&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = wc ( ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Expected: [0..9, a..z, A..Z, _] Actual: &quot; } ] } , input : &quot; Hello&quot; } = Ergo . parse ( parser , &quot; Hello&quot; )","ref":"Ergo.Terminals.html#wc/0","title":"Ergo.Terminals.wc/0","type":"function"},{"doc":"The ws/0 parser accepts a white space character and is equivalent to the \\s regular expression. Examples iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = ws ( ) iex&gt; assert % Context { status : :ok , ast : ?\\s , input : &quot;World&quot; , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot; World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = ws ( ) iex&gt; assert % Context { status : :ok , ast : ?\\t , input : &quot;World&quot; , index : 1 , line : 1 , col : 2 } = Ergo . parse ( parser , &quot; \\t World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = ws ( ) iex&gt; assert % Context { status : :ok , ast : ?\\n , input : &quot;World&quot; , index : 1 , line : 2 , col : 1 } = Ergo . parse ( parser , &quot; \\n World&quot; ) iex&gt; alias Ergo.Context iex&gt; import Ergo.Terminals iex&gt; parser = ws ( ) iex&gt; assert % Context { status : { :error , [ { :unexpected_char , &quot;Expected: [ \\s , \\t , \\r , \\n , \\v ] Actual: H&quot; } ] } , input : &quot;Hello World&quot; } = Ergo . parse ( parser , &quot;Hello World&quot; )","ref":"Ergo.Terminals.html#ws/0","title":"Ergo.Terminals.ws/0","type":"function"},{"doc":"","ref":"Ergo.Utils.html","title":"Ergo.Utils","type":"module"},{"doc":"Examples iex&gt; Ergo.Utils . char_to_string ( ?H ) &quot;H&quot; iex&gt; Ergo.Utils . char_to_string ( ?h ) &quot;h&quot; iex&gt; Ergo.Utils . char_to_string ( ?\\r ) &quot; \\r &quot; iex&gt; Ergo.Utils . char_to_string ( ?\\n ) &quot; \\n &quot; iex&gt; Ergo.Utils . char_to_string ( ?\\t ) &quot; \\t &quot;","ref":"Ergo.Utils.html#char_to_string/1","title":"Ergo.Utils.char_to_string/1","type":"function"},{"doc":"If the string str is longer than max (default: 40) it is truncated to 37 chars and &quot;...&quot; is added to the end, otherwise it is returned unaltered. Examples iex&gt; Ergo.Utils . ellipsize ( &quot;frob&quot; ) &quot;frob&quot; iex&gt; Ergo.Utils . ellipsize ( &quot;123456789012345678901234567890123456789012345678901&quot; ) &quot;1234567890123456789012345678901234567...&quot; iex&gt; Ergo.Utils . ellipsize ( &quot;12345678901234567890&quot; , 10 ) &quot;1234567...&quot; iex&gt; Ergo.Utils . ellipsize ( &quot;12345&quot; , 4 ) &quot;1...&quot;","ref":"Ergo.Utils.html#ellipsize/2","title":"Ergo.Utils.ellipsize/2","type":"function"},{"doc":"","ref":"Ergo.Parser.CycleError.html","title":"Ergo.Parser.CycleError","type":"exception"},{"doc":"Overview Ergo is a parser combinator library written in Elixir. The name 'Ergo' comes from 'ergo' which mean' 'therefore' which means 'for that reason' which seemed appropriate for a parser. If you are already familiar with parser combinators, great! If not we have a gentle introduction or see Saša Jurić's talk where he builds up parser combinators from the ground up . If you want to know why you should (or shouldn't) use Ergo see comparison with other parser combinator libraries .","ref":"overview.html","title":"Overview","type":"extras"},{"doc":"Add Ergo to your mix.exs file: { :ergo , &quot;~&gt; 0.4&quot; } Then run mix deps . get","ref":"overview.html#how-to-install-ergo","title":"Overview - How to install Ergo.","type":"extras"},{"doc":"Intro to Parser Combinators If you prefer to watch to a talk see Saša Jurić build up parser combinators from the ground up The phrase parser combinator may sound confusing but is in fact deceptively simple if you already understand the notion of functions and function composition. For example transforming [1, 2, 3, 4, 5] into [2, 4, 6, 8, 10] can be achieved through combining the functions Enum.map and double as follows: double = fn x -&gt; 2 * x end Enum . map ( [ 1 , 2 , 3 , 4 , 5 ] , double ) If you think of a parser as being a certain kind of function that works on a textual input (in the way that double above is a kind of function that works on a number input) then you may see how we can combine parsers to do work. For example to parse an input such as &quot;12345&quot; we might combine two parsers many and digit . Let's imagine digit is a parser that accepts an input character in the range of '0'..'9' and returns a number, like this: digit ( &quot;12345&quot; ) = 1 digit ( &quot;2345&quot; ) = 2 and so on. Now lets imagine that many is a parser that accepts an input and another parser and keeps attempting to match that parser against the input until it doesn't match any further. So, for example: many ( &quot;12345&quot; , digit ) = [ 1 , 2 , 3 , 4 , 5 ] It's the &quot;accepts another parser&quot; concept that is key and is what makes many a combinator parser. In the same way as map calls the function double to do work, many calls the parser digit to do work. The insight is that we can replace digit with any other parser including other combinator parsers. In this way we can combine, and recombine, simpler parsers to form more complex parsers until we have something that can parse our input. In Ergo we distinguish between terminal parsers that only work on the input and combinator parsers that are parameterised with one or more other parsers (which may themselves be either a terminal or another combinator parser).","ref":"parser_combinator_intro.html","title":"Intro to Parser Combinators","type":"extras"},{"doc":"Comparisons with other libraries There are already at least two mature parser combinator libraries for Elixir: NimbleParsec and ExSpirit so why write another one? First I am learning about writing parsers with parser combinators and implementing my own seemed like a good way to learn more. I was especially motivated by Saša Jurić's talk where he builds up parser combinators from the ground up . Ergo owes quite a lot to his style. Second the code of both ExSpirit &amp; NimbleParsec is, given my relatively limited experience of Elixir, a little hard to digest. I wanted to handle errors better than I'd seen in any of their examples (which seemed to ignore this issue) and finding the code hard to understand made it difficult to do that. So there are some notable differences between Ergo and NP/ExS that perhaps justify its existence: Firstly, and perhaps most difficult to notice, is that Ergo is implemented in terms of functions only, while ExS/NP both make significant use of macros. My understanding is not sophisticated enough to understand quite what the macros are buying the user but I was able to do without them. But I find the Ergo code relatively easy to read after writing it. Second Ergo is built with error handling as one of its priorities. Internally, Ergo parsers are not bare functions but Parser structs that assist with debugging. This takes the form of logging parser operations and cycle detection. See further down for more information. Lastly, ExS/NP make heavy use of the Elixir |&gt; operator to combine parsers together, for example: act = string ( &quot;Act&quot; ) |&gt; replace ( :act ) |&gt; ignore ( whitespace ) |&gt; concat ( id ) |&gt; ignore ( whitespace ) |&gt; ignore ( char ( ?{ ) ) |&gt; ignore ( whitespace ) |&gt; optional ( attributes ) |&gt; ignore ( whitespace ) |&gt; concat ( scene ) |&gt; repeat ( ignore ( whitespace ) |&gt; concat ( scene ) ) |&gt; ignore ( whitespace ) |&gt; ignore ( char ( ?} ) ) |&gt; wrap while in Ergo style you'd write: act = sequence ( [ literal ( &quot;Act&quot; , ast : fn _ast -&gt; :act end ) , ignore ( whitespace ) , id , ignore ( whitespace ) , ignore ( char ( ?) ) , ignore ( whitespace ) , optional ( attribtues ) , ignore ( whitespace ) , many ( scene ) , ignore ( whitespace ) , ignore ( char ( ?} ) ) ] ) Perhaps unsurprisingly I prefer the Ergo style. That said, if you have serious parsing intentions you should probably be using one of NimbleParsec or ExSpirit instead of Ergo. Ergo is likely less complete, less reliable, and certainly less performant than either of those more established libraries.","ref":"comparisons.html","title":"Comparisons with other libraries","type":"extras"},{"doc":"Creating a Basic Parser This tutorial will explore how Ergo works through creating a parser to match numbers, be they integers like -5 or 42, or decimals like 25.6. The parser will match them and convert them into the appropriate Elixir numeric value. We will build it up in stages, starting with bare digits.","ref":"basic_parser.html","title":"Creating a Basic Parser","type":"extras"},{"doc":"To begin with we need to be able to parse digits. One option for that is to use the basic char parser to match digit characters. Using IEx here is what you would do: alias Ergo alias Ergo.Context import Ergo.Terminals digit = char ( ?0 .. ?9 ) Ergo . parse ( digit , &quot;42&quot; ) % Context { status : :ok , ast : 52 } Where the integer 52 is the character code of the digit '4' (Try typing ?4 into the IEx console to see for yourself). Now we can parse a single digit, how about multiple digits?","ref":"basic_parser.html#parsing-a-digit","title":"Creating a Basic Parser - Parsing a digit","type":"extras"},{"doc":"To parse multiple digits we use the many parser in conjunction with the digit parser as follows: import Ergo.Combinators digits = many ( digit ( ) ) Ergo . parse ( digits , &quot;42&quot; ) % Context { status : :ok , ast : [ 52 , 50 ] } In fact you might see '42' in your version of the AST because IEx will try to render the list [52, 50] as a charlist. This is a hangover from Erlang. If you would prefer to see the list add the following to your ~/.iex.exs file: IEx . configure ( inspect : [ charlists : :as_lists ] ) The many combinator parser repeatedly invokes the digit parser to match as many digit characters as possible, generating an AST that is a list of those digits. To transform them into a numeric value we need to apply another function to the AST. In our case the AST list contains character values of the digits matched, e.g. [52, 50] for the digits ['4', '2'] respectively. Since the digit character '0' has character value of 48 we can turn the characters into digit values by subtracting 48. The heavy lifting of transforming digits is done by c_transform below which is pipeline to transform ['4', '2'] -&gt; [{4, 10}, {2, 1}] -&gt; [40, 2] -&gt; 42: c_transform = fn ast -&gt; bases = Stream . unfold ( 1 , fn n -&gt; { n , n * 10 } end ) digits = Enum . map ( ast , fn digit -&gt; digit - 48 end ) Enum . zip ( Enum . reverse ( digits ) , bases ) |&gt; Enum . map ( &amp; Tuple . product / 1 ) |&gt; Enum . sum end digits = many ( digit ) |&gt; transform ( c_transform ) Ergo . parse ( digits , &quot;42&quot; ) % Context { status : :ok , ast : 42 } In this case we are applying the transform parser to the many parser. Transform only operates on the AST of the parser it is given by applying a function to it. In this case we could also have used: digits = many ( digit ( ) , ast : c_transform ) As many of the combinator parsers support an optional ctx: or ast: argument as a shortcut. At this point we can parse positive integers of any length: Ergo . parse ( digits , &quot;918212812783918723&quot; ) % Context { status : :ok , ast : 918212812783918723 }","ref":"basic_parser.html#parsing-many-digits","title":"Creating a Basic Parser - Parsing many digits","type":"extras"},{"doc":"What about negative values? We need to look for a leading '-' character however, unlike the digits, the minus is optional. Parsing the minus is simple enough: minus = char ( ?- ) Ergo . parse ( minus , &quot;-&quot; ) % Context { status : :ok , ast : 45 } 45 is the char value of the char '-'. We can now use the optional combinator to allow a minus to be matched, or not: minus = optional ( char ( ?- ) ) Ergo . parse ( minus , &quot;-42&quot; ) % Context { status : :ok , ast : 45 } Ergo . parse ( minus , &quot;42&quot; ) % Context { status : :ok , ast : nil } In the second case the status is :ok meaning the optional parser succeeded, however the ast is nil meaning nothing was matched. Let's make this a bit more useful: minus = optional ( char ( ?- ) ) |&gt; transform ( fn ast -&gt; case ast do nil -&gt; 1 45 -&gt; - 1 end end ) Ergo . parse ( minus , &quot;-42&quot; ) % Context { status : :ok , ast : - 1 } Ergo . parse ( minus , &quot;42&quot; ) % Context { status : :ok , ast : 1 } Now when minus matches a '-' it will transform it to the value -1. When it doesn't match anything it will transform it to the value 1. Now let's combine it with the other parser. integer = sequence ( [ minus , digits ] ) The sequence parser tries to match a list of parser in turn and, if they all match, generates an AST composed of a list of each of their results. Let's see how it works: Ergo . parse ( integer , &quot;1234&quot; ) % Context { status : :ok , ast : [ 1 , 1234 ] } Ergo . parse ( integer , &quot;-5678&quot; ) % Context { status : :ok , ast : [ - 1 , 5678 ] } So we can see that it's easy to get the right result by simply taking the product of the two values in the AST: integer = sequence ( [ minus , digits , ] , ast : &amp; Enum . product / 1 ) Ergo . parse ( integer , &quot;1234&quot; ) % Context { status : :ok , ast : 1234 } Ergo . parse ( integer , &quot;-5678&quot; ) % Context { status : :ok , ast : - 5678 } So far so good. We can now parse positive and negative integers.","ref":"basic_parser.html#parsing-negative-numbers","title":"Creating a Basic Parser - Parsing negative numbers","type":"extras"},{"doc":"If we want to parse decimal numbers as well we need to handle the (optional) mantissa, the digits to the right of the decimal point. We can see that the mantissa is structurally the same as the integer part, a set of digits, but will need to be processed a little differently. The m_transform function below should look familiar. It works the same way as the c_transform only instead of multiplying by increasing powers of 10, we're dividing by increasing powers of 10. m_transform = fn ast -&gt; ast |&gt; Enum . map ( fn digit -&gt; digit - 48 end ) |&gt; Enum . zip ( Stream . unfold ( 0.1 , fn n -&gt; { n , n / 10 } end ) ) |&gt; Enum . map ( &amp; Tuple . product / 1 ) |&gt; Enum . sum end mantissa = many ( digit , ast : m_transform ) Ergo . parse ( mantissa , &quot;5&quot; ) % Context { status : :ok , ast : 0.5 } Ergo . parse ( mantissa , &quot;42&quot; ) % Context { status : :ok , ast : 0.42000000000000004 } There may be a precision issue with this code but you can see the principle it is operating by. Now to join the two components together, assuming there is a decimal point (suggesting we'll need optional again). Also we'll again make use of the ast: feature of the sequence combinator to process AST's to give us the right value. number = sequence ( [ integer , optional ( sequence ( [ ignore ( char ( ?. ) ) , mantissa ] , ast : &amp; List . first / 1 ) ) ] , ast : &amp; Enum . sum / 1 ) Ergo . parse ( number , &quot;42&quot; ) % Context { status : :ok , ast : 42 } Ergo . parse ( number , &quot;0.45&quot; ) % Context { status : :ok , ast : 0.45 } Ergo . parse ( number , &quot;-42&quot; ) % Context { status : :ok , ast : - 42 } All looking good, just one more example: Ergo . parse ( number , &quot;-4.2&quot; ) % Context { status : :ok , ast : - 3.8 } Oops! There is a problem with our implementation in that we add together the integer and decimal parts. This works for positive numbers but in the latter case -4 + 0.2 = -3.8 not -4.2. When the integer part is negative we need to subtract the decimal part. We can no longer just use Enum.sum to process the result of the top-level sequence. Instead: combine = fn [ integer , decimal | [ ] ] -&gt; if integer &gt;= 0 do integer + decimal else integer - decimal end ast -&gt; Enum . sum ( ast ) end number = sequence ( [ integer , optional ( sequence ( [ ignore ( char ( ?. ) ) , mantissa ] , ast : &amp; List . first / 1 ) ) ] , ast : combine ) Ergo . parse ( number , &quot;-4.2&quot; ) % Context { status : :ok , ast : - 4.2 }","ref":"basic_parser.html#parsing-decimals","title":"Creating a Basic Parser - Parsing decimals","type":"extras"},{"doc":"Through a series of steps we have built a parser that can handle any kind of integer or decimal number we throw at it. We've seen the using of terminal parsers like char as well as combinator parsers like optional , ignore , many , and sequence and meta parsers like transform (and that often transform can be specified as a ast: argument to a combinator parser). Hopefully this guide will be helpful in thinking about how to build your own parsers.","ref":"basic_parser.html#conclusion","title":"Creating a Basic Parser - Conclusion","type":"extras"},{"doc":"Debugging ** THIS SECTION IS OUT OF DATE. See also Ergo.diagnose . As parsers become more complex it can be difficult to work out why they fail to work properly for a given input. Ergo parsers can, by default, generate debugging information to tell you about what is happening and what decisions are being made. To use this feature call the parse function as follows: Ergo . parse ( parser , input , debug : true ) The various parsers will now use Logger.info to record information about how they are processing their inputs. A challenge when building parsers is accidentally creating a cycle where the parser will never finish but loop over the same input forever. For example: many ( choice ( [ many ( ws ( ) ) , char ( ?} ) ] ) ) Given an input like &quot;}}}&quot; will never finish. The inner many clause will always succeed with 0 whitespace characters and never actually process the char parser at all. You wouldn't deliberately set out to write such a parser but it can happen or at least it seems to happen to me. For this reason Ergo implements cycle detection. Parsers with the track: true option (which includes most of the combinator parsers) record in the context when they have run and the current index into the input. If the same parser is run a second time on the same input we know we have hit a cycle and an Ergo.Context.CycleError will be raised.","ref":"debugging.html","title":"Debugging","type":"extras"},{"doc":"Recursion","ref":"recursion.html","title":"Recursion","type":"extras"},{"doc":"Parser combinators, being a form of recursive descent parser, are unable to handle left recursion. In general this looks like: A → A𝛼 | β Where non-terminal A ends up being substituted for itself before any token can be matched, leading to A → A𝛼 | β again, and so on and so on in an infinite recursion. Such parsers can usually be rewritten to a form that is not left-recursive however that is beyond the scope of this guide.","ref":"recursion.html#left-recursion","title":"Recursion - Left Recursion","type":"extras"},{"doc":"There is a further problem that arises from the way parser combinators are defined as functions returning functions. Here is an example parser that is designed to parse values like 42, true, &quot;What is six times seven?&quot; and also lists of values, including other lists. Note that we elide handling white space for brevity. def value ( ) do choice ( [ number_val ( ) , string_val ( ) , boolean_val ( ) , list_val ( ) ] ) end def list_val ( ) do sequence ( [ char ( ?[ ) value ( ) , many ( sequence ( [ comma ( ) , value ( ) ] ) ) , char ( ?] ) ] ) end A value can be a number, string, boolean, or list. But list is a sequence of values. The problem arises from a call to value() leading to a call to list() which, in turn, leads to a call to value(), leading to a call to list() and so on in an infinite recursion. Note that this is not grammatical recursion as in the left recursion example above and cannot be solved by rewriting. The issue is that Elixir evaluates function calls eagerly, i.e. when they are encountered. We need to introduce a gap to break the recursion. We do this with the lazy parser combinator. Lazy wraps it's parser in a function so that it is not called immediately, breaking the recursion. E.g.: def value ( ) do choice ( [ number_val ( ) , string_val ( ) , boolean_val ( ) , lazy ( list_val ( ) ) ] ) end","ref":"recursion.html#eager-recursion","title":"Recursion - Eager Recursion","type":"extras"},{"doc":"General Guidance","ref":"guidance.html","title":"General Guidance","type":"extras"},{"doc":"In my experience when you already have an idea of what it is you want to parse there is a temptation to start at the top and work down. Whenever I do this I seem to run into trouble with something that doesn't fit together right or parsing issues that I then find hard to track down. My experience is that it's easiest to work bottom up. Parse small actual pieces of the language you want to parse and join these together, building towards the whole format. In general I tend to find I hit parsing errors quicker and can resolve them more easily this way. While you might have a different experience my advice, if you don't have an opinion yet, is to build your parser bottom up.","ref":"guidance.html#top-down-or-bottom-up","title":"General Guidance - Top-down or Bottom-up?","type":"extras"}]